"""Utilities for interacting with our cloud db (Deta Base - not a typo).
"""
import builtins
from collections import deque
import pandas as pd
from datetime import datetime
from deta import Deta
import json
from pathlib import Path
import warnings

# pd_tools makes coalesce available
from htools import tolist, pd_tools, random_str
from maminbot import config as cfg


def collapse_related_docs(chunk, concat_col='text', sep='\n'):
    if chunk.shape[0] <= 1:
        return chunk.reset_index(drop=True)
    nested_row = {
        col: sep.join(chunk[col].tolist()) if col == concat_col
        else chunk[col].values[0]
        for col in chunk.columns
    }
    # Don't use from_dict because that converts lists of 1 back into
    # primitives.
    return pd.DataFrame([nested_row])


def fetch_records(user='', tags=(), tag_strategy='any', prompt_contains='',
                  text_contains='', db=None, collapse=True, **kwargs):
    """Fetch records from Deta Base maminbotDB.

    Parameters
    ----------
    user: str
        Optionally provide one of the names in maminbot.config.users to only
        retrieve text from that user.
    tags: list[str]
        Optionally provide one or more tags to look for. Depending on tag
        strategy, you can choose to keep only rows that contain all tags or
        keep rows that contain any of the specified tags.
    tag_strategy: str
        "any" or "all". "any" returns rows that contain any of the tags the
        user passed in, while "all" requires rows to possess all specified tags
        in order to be returend.
    prompt_contains: str
        If provided, we only return rows where the "prompt" columnn contains
        the specified text. Case insensitive. Rows with no prompt will never be
        returned.
    text_contains: str
        If provided, we only return rows where the "text" columnn contains the
        specified text. Case insensitive.
    db: deta.base._Base
        Deta Base object. Typically created by
        `deta.Deta(project_key).Base(project_name)` where project_name is
        "maminbotDB".
    collapse: bool
        If True, any docs that were stored from a single piece of input text
        will be combined back into one row. Notice how in store_record in this
        same module, we have to do this for long docs (>50k characters) to
        avoid Deta http errors.
    kwargs: any
        Ignored - just makes it easier to pass in all DataTrainingArguments
        in s00 and maminbot.datasets.LanguageModelDataset.

    Returns
    -------
    pd.DataFrame
    """
    if not db:
        db = connect_db()
    query = {}
    if user:
        query['user'] = user
    tags = tolist(tags)
    if tags and tag_strategy not in ('any', 'all'):
        raise ValueError(f'Got unexpected tag_strategy {tag_strategy}. '
                         f'Should be "any" or "all".')
    res = db.fetch(query).items
    df = pd.DataFrame(res)
    if tags:
        func = getattr(builtins, tag_strategy)
        df = df[df.tags.apply(lambda x: func(tag in x for tag in tags))]
    if prompt_contains:
        df = df[df.prompt.str.lower().str.contains(prompt_contains.lower())]
    if text_contains:
        df = df[df.text.str.lower().str.contains(text_contains.lower())]
    df['dt'] = pd.to_datetime(df.dt)
    if collapse and 'doc_id' in df.columns:
        df['doc_id'] = df.coalesce(['doc_id', 'key'])
        cols = df.columns.drop('doc_id').tolist()
        df = df.sort_values('dt', ascending=True)\
            .groupby('doc_id')[cols]\
            .apply(collapse_related_docs)
    return df.reset_index(drop=True)


def load_creds(secrets=None):
    """
    Output isn't actually mutated but streamlit always raises a warning here
    unless we allow output mutation.
    NOTE: debug must be after st.cache. Do NOT use htools.decorate_functions
    because then debug will be on the outside and caching won't work.
    """
    if secrets:
        return dict(secrets.deta_creds)
    try:
        df = pd.read_csv(cfg.root/'data/creds/deta.csv')
        return df.to_dict(orient='records')[0]
    except FileNotFoundError:
        # This is how we load creds on saturn cloud. We made this available
        # by calling load_creds locally, passing that to json.dumps() and
        # copy-pasting it into a saturn secret. Then we attached that secret
        # to the relevant gpu resource as a file at the path below.
        with open(Path('~/secrets/deta_creds').expanduser(), 'r') as f:
            return json.load(f)


def connect_db(project_key='', project_name='', **kwargs):
    """Load deta.Base object that can be used to fetch items from our cloud
    database.

    Parameters
    ----------
    project_key: str
        Deta "project_key". Should not expose this on github.
    project_name: str
        Deta "project_name". This doesn't need to be private.
    kwargs: any
        Extra kwargs are ignored. Just makes it so streamlit secrets is less
        brittle - we can include all deta creds without worrying about which
        args are necessary.

    Returns
    -------
    deta.base._Base
    """
    if not (project_key and project_name):
        kwargs.update(DETA_CREDS)
        project_key = kwargs['project_key']
        project_name = kwargs['project_name']
        warnings.warn(f'Missing key and/or name so defaulting to '
                      f'{project_name} project creds.')
    for obj in (project_key, project_name):
        if not isinstance(obj, str) or obj[0] == '{':
            raise ValueError(
                'Input args should be strings. It looks like you might have '
                f'passed in a dict or stringified dict by accident: {obj}'
            )
    deta = Deta(project_key)
    return deta.Base(project_name)


def store_record(db, user, prompt_type, prompt, tags, text, _internal=False,
                 doc_id=''):
    """Add a record to the maminbotDB Deta Base, chunking it up into multiple
    docs if necessary since Deta Base errors out once we hit ~70k characters.

    Parameters
    ----------
    db: deta.base._Base
        Obj returned by connect_db(), for example.
    user: str
        One of maminbot.config.users, e.g. 'adrienne'. This will be
        auto-lowercased.
    prompt_type: str
        Dev-friendly prompt name, usually a gpt task name. One of
        ("misc_writing_prompt_ideas",
         "journal_entry_ideas",
         "creative_writing_ideas").
    prompt: str
        Text prompt that inspired the written text. 'n/a' will be considered
        to be no prompt (stored as empty str).
    tags: list[str]
        Let users provide tags to make it easier to train task-specific models
        later.
    text: str
        The user-provided piece of writing.
    _internal: bool
        User should never pass this in explicitly. The function uses this to
        differentiate between user calls and recursive calls, which occur when
        the provided text is too long to store in a single Deta doc.
    doc_id: str
        User shouldn't specify this manually. It helps us store very long
        documents in multiple Deta docs (because they produce an http error
        otherwise) that can later be reconstructed into the original doc if
        desired.

    Returns
    -------
    list[dict]: each dict corresponds to one doc that was added to the db.
    This includes a newly generated PK 'key' field. A long piece of input text
    can end up generating many docs, but anything <50k characters will be kept
    in a single doc.
    """
    doc_id = doc_id or random_str(10)
    user = user.lower()
    if prompt.lower() == 'n/a':
        prompt = ''
        prompt_type = ''
    max_chars = 50_000
    length = len(text)
    if not _internal and length > max_chars:
        warnings.warn(
            'Chunking up text into smaller documents so we can add them to '
            'the db.'
        )
        lines = deque(text.splitlines())
        all_res = []
        while lines:
            cur_len = 0
            cur = []
            while cur_len < max_chars and lines:
                line = lines.popleft()
                cur.append(line)
                cur_len += len(line)
            cur_res = store_record(db, user, prompt_type, prompt, tags,
                                   '\n'.join(cur), _internal=True,
                                   doc_id=doc_id)
            all_res += cur_res
        return all_res

    record = {
        'dt': str(datetime.now()),
        'text': text,
        'prompt_type': prompt_type,
        'prompt': prompt,
        'tags': tags,
        'user': user,
        'doc_id': doc_id,
    }
    res = db.put(record)
    return [res]


try:
    DETA_CREDS = load_creds()
except Exception as e:
    warnings.warn('Deta credentials could not be loaded. DETA_CREDS will be '
                  'initialized to an empty dict.\n' + str(e))
    DETA_CREDS = {}



from itertools import cycle
import numpy as np
import torch
from torch.utils.data import Dataset
from tqdm.auto import tqdm
import warnings

from maminbot.db_utils import fetch_records
from maminbot.utils import dataclass_to_dict


class LanguageModelDataset(Dataset):

    def __init__(self, tokenizer, texts=None, x=None,
                 fixed_len=True, drop_last=False, **config):
        """
        Parameters
        ----------
        tokenizer
        texts: List[str] or None
            List of pieces of text. Provide either this OR `texts`.
        x: List[dict] or None
            List of dicts, each containing at least an "input_ids" key.
            This is the result of one of our chunkify methods.
            Provide either this OR `texts`. Useful for creating subsets or
            train/test splits.
        fixed_len: bool
            If True, make all token sequences the same length (model max
            length). This lets us avoid using padding and can dramatically
            speed up training.
        drop_last: bool
            Only used when fixed_len=True. Drops the last sequence if it's too
            short because otherwise the sequences wouldn't all be the
            same length.
        config: any
            Specified automatically when using from_data_args method, which is
            used in s00 to train a model from the command line. It's just a
            dict containing all the DataTrainingArguments provided by the user.
        """
        if tokenizer.eos_token_id is None:
            raise ValueError('Tokenizer eos_token_id cannot be None.')

        self.tokenizer = tokenizer
        config.update(fixed_len=fixed_len, drop_last=drop_last)
        # x will be a list of dicts. Input validation must remain after the two
        # lines above (or pass in tokenizer and other inputs too).
        self._validate_inputs(x, texts, config)
        self.x = x or self.chunkify(texts, fixed_len=fixed_len,
                                    drop_last=drop_last)
        self.config = config

    def _validate_inputs(self, x, texts, config):
        """Because this class allows passing in either raw text (list[str]) or
        pre-defined dataset items (dict containing tensor(s)) AND we support
        a couple different modes (dataset items may or may not contain
        attention masks, depending on choice of fixed_len in init), we need to
        do a lot of validation of input args.

        Parameters
        ----------
        x: list[dict] or None
        texts: list[str] or None
        config: dict
        """
        if not x:
            # Can skip most validation because we'll be constructing the inputs
            # ourselves.
            assert texts, 'Texts must be provided when x is null.'
            return

        if texts:
            warnings.warn('Texts will be ignored because x is not None.')
        first = list(x[0])
        if config['fixed_len']:
            # Last item may be different length when drop_last is True, so we
            # discard it in that case (only for validation purposes - we're
            # just changing the local variable here).
            x = x[:len(x) - config['drop_last']]
            lengths = set(len(row['input_ids']) for row in x)
            assert len(lengths) == 1, 'Found mismatched lengths when ' \
                                      'fixed_len=True.'
            assert next(iter(lengths)) == self.tokenizer.model_max_length, \
                'Sequence length should equal model max length when ' \
                'fixed_len=True.'
            expected_first = ['input_ids']
            assert first == expected_first, \
                f'Each dataset item should only have keys {expected_first} ' \
                f'when fixed_len=True, but found: {first}'
        else:
            expected_first = ['attention_mask', 'input_ids']
            first = sorted(first)
            assert first == ['attention_mask', 'input_ids'], \
                f'Each dataset item should have keys {expected_first} ' \
                f'when fixed_len=False. Instead found {first}.'

    def chunkify(self, texts, fixed_len, drop_last):
        """Chunk up texts into dataset items (dicts of tensor(s)).

        Parameters
        ----------
        texts: list[str]
        fixed_len: bool
        drop_last: bool

        Returns
        -------
        list[dict]
        """
        if fixed_len:
            chunks = self._chunkify_fixed_len(texts, drop_last=drop_last)
        else:
            chunks = self._chunkify_variable_length(texts)
        return list(chunks)

    def _chunkify_variable_length(self, texts):
        # This method chunks up each row separately such that each chunk
        # contains only 1 source sentence.
        warnings.warn('We recommend setting fixed_len=True to speed up '
                      'training. This implementation is far slower.')

        # Fewer dot accesses in loop speeds things up when dataset is huge.
        tokenizer = self.tokenizer
        chunk_size = tokenizer.model_max_length
        for row in tqdm(texts, desc='Tokenizing corpus...'):
            tokens = tokenizer(row, return_tensors='pt')
            chunks = [{'input_ids': chunk.squeeze(0),
                       'attention_mask': torch.ones(chunk.shape[-1])}
                      for chunk
                      in tokens['input_ids'].split(chunk_size, dim=-1)]
            yield from chunks

    def _chunkify_fixed_len(self, texts, drop_last=True):
        # Chunks dataset up into fixed length chunks so we don't require
        # padding or masking during training.
        # Fewer dot accesses in loop speeds things up when dataset is huge.
        tokenizer = self.tokenizer
        chunksize = tokenizer.model_max_length
        if not isinstance(texts, list):
            texts = list(texts)
        flat = []
        for row in tqdm(texts):
            flat.extend(
                tokenizer(row, truncation=False, padding=False).input_ids
                + [tokenizer.eos_token_id]
            )
        total_tokens = len(flat)
        for i in range(0, total_tokens, chunksize):
            chunk = flat[i:i + chunksize]
            # When we train on a tiny subset, we could end up with less than 1
            # full max len sequence, so we make sure we never drop the first
            # chunk or we could end up with a ds of length 0.
            cur_length = len(chunk)
            if cur_length < chunksize and i > 0 and drop_last:
                warnings.warn(f'Dropping last chunk with length {len(chunk)}:'
                              f'\n\n{tokenizer.decode(chunk)}')
                break
            # If our first sequence is shorter than the max, we don't need to
            # pad it because all our sequences will be the same length
            # regardless. That should really only happen when testing on tiny
            # toy datasets.
            if i > 0:
                chunk = pad_with_self(chunk, chunksize, tokenizer.eos_token_id)
            yield {'input_ids': torch.tensor(chunk, dtype=torch.long)}

    @classmethod
    def from_data_args(cls, tokenizer, data_args):
        """Create dataset from DataArguments in s00 (i.e. when we train a model
        from the command line).

        Parameters
        ----------
        tokenizer
        data_args: Union[dict, DataTrainingArguments]
        """
        kwargs = dataclass_to_dict(data_args)
        df = fetch_records(**kwargs)
        ds = cls(tokenizer, texts=df.text.tolist(), **kwargs)
        if kwargs.get('fast_dev_run'):
            # One training batch, one validation batch.
            n_seq = kwargs.get('batch_size', 1) * 2
            ds = ds.subset(n_seq)
        return ds

    def train_test_split(self, test_size, seed=0):
        """
        Parameters
        ----------
        test_size: int or float
            If float, should be in [0, 1). If int, should be in [0, len(self)).
        seed: int

        Returns
        -------
        dict[str, LanguageModelDataset]: Key 'train' is always present. Key
        'test' is only present if test_size > 0.
        """
        if not test_size:
            return {'train': self}
        length = len(self)
        if isinstance(test_size, float):
            test_size = int(test_size * length)
        if test_size >= length:
            raise ValueError(
                'Your requested test set size is too big. You only have '
                f'{length} sequences available for training and testing '
                f'combined.'
            )
        test_idx = set(
            np.random.RandomState(seed).choice(
                length, test_size, replace=False
            )
        )
        train_x = []
        test_x = []
        for i, row in enumerate(self.x):
            if i in test_idx:
                test_x.append(row)
            else:
                train_x.append(row)
        cls = type(self)
        ds_train = cls(self.tokenizer, x=train_x, **self.config)
        ds_test = cls(self.tokenizer, x=test_x, **self.config)
        return {'train': ds_train,
                'test': ds_test}

    def subset(self, n_sequences=1, strict=False):
        """Create a subset of an existing LanguageModelDataset.

        Parameters
        ----------
        n_sequences
        strict

        Returns
        -------
        LanguageModelDataset
        """
        if n_sequences > len(self):
            msg = (f'Cannot create a subset with {n_sequences} '
                   f'sequences when the original datset only has '
                   f'{len(self)}.')
            if strict:
                raise ValueError(msg)
            else:
                warnings.warn(msg)
        cls = type(self)
        return cls(self.tokenizer, x=self.x[:n_sequences], **self.config)

    def __len__(self):
        return len(self.x)

    def __getitem__(self, i):
        return self.x[i]


def cycle_chunks(arr, chunksize, sep):
    """Like itertools.cycle but we yield k items at a time instead of 1
    and we insert a separator item between the last and first item in the
    sequence when we loop around. Used to pad out a batch of text when training
    with fixed sequence lengths (hate to throw away data when it's scarce).

    Parameters
    ----------
    arr: list
    chunksize: int
        Desired size of each chunk. The sep char counts as 1, the same as each
        item in the original list.
    sep: any
        Typically the same type as the items in arr.

    Returns
    -------
    list

    Examples
    --------
    ```
    chunksize = 7
    for i, row in enumerate(iter_chunks([0, 1, 2, 3, 4], chunksize, '<EOS>')):
        print(row)
        assert len(row) == chunksize
        # Exit at arbitrary point to prevent infinite loop.
        if i >= 10: break
    ```

    [0, 1, 2, 3, 4, '<EOS>', 0]
    [1, 2, 3, 4, '<EOS>', 0, 1]
    [2, 3, 4, '<EOS>', 0, 1, 2]
    [3, 4, '<EOS>', 0, 1, 2, 3]
    [4, '<EOS>', 0, 1, 2, 3, 4]
    ['<EOS>', 0, 1, 2, 3, 4, '<EOS>']
    [0, 1, 2, 3, 4, '<EOS>', 0]
    [1, 2, 3, 4, '<EOS>', 0, 1]
    [2, 3, 4, '<EOS>', 0, 1, 2]
    [3, 4, '<EOS>', 0, 1, 2, 3]
    [4, '<EOS>', 0, 1, 2, 3, 4]
    """
    arr = list(arr) + [sep]
    generator = cycle(arr)
    while True:
        chunk = []
        for i, x in enumerate(generator, 1):
            chunk.append(x)
            if i == chunksize: break
            i += 1
        yield chunk


def pad_with_self(arr, size, sep):
    """Pad out a sequence to a desired size by looping around on itself.
    See cycle_chunks in the same module - this just takes the first one.

    Parameters
    ----------
    arr: list
        Typically a list of ints (token IDs).
    size: int
        Desired size.
    sep: any
        Usually an int. Should typically be the same type as the items in arr.

    Returns
    -------
    list: Has same input type as the items in arr (assuming sep has same type).
    If arr is already the desired size, we return itself. If it's longer than
    the desired size, we return arr[:size].
    """
    res = next(cycle_chunks(arr, size, sep))
    assert len(res) == size, f'Expected len={size} but got len={len(res)}.'
    return res



from pathlib import Path
from torch import nn
from transformers.generation_logits_process import LogitsProcessorList
from transformers.generation_stopping_criteria import (
    StoppingCriteriaList,
    validate_stopping_criteria,
)
from transformers.generation_utils import GreedySearchEncoderDecoderOutput, \
    GreedySearchDecoderOnlyOutput, SampleEncoderDecoderOutput, \
    SampleDecoderOnlyOutput
from transformers.utils import logging
from typing import Iterable, List, Optional, Tuple, Union
import inspect
import torch
import torch.distributed as dist
import warnings

from htools import tolist, save, spacer, add_docstring


logger = logging.get_logger(__name__)
GreedySearchOutput = Union[GreedySearchEncoderDecoderOutput,
                           GreedySearchDecoderOnlyOutput]
SampleOutput = Union[SampleEncoderDecoderOutput, SampleDecoderOnlyOutput]
GenerateOutput = Union[GreedySearchOutput, SampleOutput]


def ids_to_pretty_tokens(tokenizer, ids):
    """Similar to tokenizer.decode but returns a list of strings rather than a
    single string. Similar to tokenizer.convert_ids_to_tokens but returns
    human-friendly tokens rather than machine-friendly (e.g. using actual
    spaces rather than some coded representation.)

    Parameters
    ----------
    tokenizer
    ids: list[int] or int

    Returns
    -------
    list[str]
    """
    return [tokenizer.convert_tokens_to_string(tok) for tok in
            tokenizer.convert_ids_to_tokens(tolist(ids))]


def _top_word_probs(probs, next_tokens, tokenizer, probs_per_token=5):
    """
    Returns
    -------
    tuple[str, dict]: String is next predicted token. w2p is a dict mapping
    token string to predicted probability.
    """
    topk = torch.topk(probs, probs_per_token)
    # Ensure actual sampled token is included.
    # Shape: (num_return_sequences, probs_per_word + 1)
    ids = torch.cat((next_tokens[:, None], topk.indices), -1)
    id_probs = torch.gather(probs, -1, ids)
    w2p = [dict(zip((ids_to_pretty_tokens(tokenizer, id_row)),
                    prob_row.tolist()))
           for id_row, prob_row in zip(ids, id_probs)]
    return ids_to_pretty_tokens(tokenizer, next_tokens), w2p


def live_greedy_search(
        model,
        tokenizer,
        input_ids: torch.LongTensor,
        logits_processor: Optional[LogitsProcessorList] = None,
        stopping_criteria: Optional[StoppingCriteriaList] = None,
        max_length: Optional[int] = None,
        pad_token_id: Optional[int] = None,
        eos_token_id: Optional[int] = None,
        output_attentions: Optional[bool] = None,
        output_hidden_states: Optional[bool] = None,
        output_scores: Optional[bool] = None,
        return_dict_in_generate: Optional[bool] = None,
        synced_gpus: Optional[bool] = False,
        probs_per_token: Optional[int] = 5,
        **model_kwargs,
) -> Union[GreedySearchOutput, torch.LongTensor]:
    r"""
    Generates sequences of token ids for models with a language modeling head using **greedy decoding** and can be
    used for text-decoder, text-to-text, speech-to-text, and vision-to-text models.
    Parameters:
        input_ids (`torch.LongTensor` of shape `(batch_size, sequence_length)`):
            The sequence used as a prompt for the generation.
        logits_processor (`LogitsProcessorList`, *optional*):
            An instance of [`LogitsProcessorList`]. List of instances of class derived from [`LogitsProcessor`]
            used to modify the prediction scores of the language modeling head applied at each generation step.
        stopping_criteria (`StoppingCriteriaList`, *optional*):
            An instance of [`StoppingCriteriaList`]. List of instances of class derived from [`StoppingCriteria`]
            used to tell if the generation loop should stop.
        max_length (`int`, *optional*, defaults to 20):
            **DEPRECATED**. Use `logits_processor` or `stopping_criteria` directly to cap the number of generated
            tokens. The maximum length of the sequence to be generated.
        pad_token_id (`int`, *optional*):
            The id of the *padding* token.
        eos_token_id (`int`, *optional*):
            The id of the *end-of-sequence* token.
        output_attentions (`bool`, *optional*, defaults to `False`):
            Whether or not to return the attentions tensors of all attention layers. See `attentions` under
            returned tensors for more details.
        output_hidden_states (`bool`, *optional*, defaults to `False`):
            Whether or not to return the hidden states of all layers. See `hidden_states` under returned tensors
            for more details.
        output_scores (`bool`, *optional*, defaults to `False`):
            Whether or not to return the prediction scores. See `scores` under returned tensors for more details.
        return_dict_in_generate (`bool`, *optional*, defaults to `False`):
            Whether or not to return a [`~utils.ModelOutput`] instead of a plain tuple.
        synced_gpus (`bool`, *optional*, defaults to `False`):
            Whether to continue running the while loop until max_length (needed for ZeRO stage 3)
        model_kwargs:
            Additional model specific keyword arguments will be forwarded to the `forward` function of the model.
            If model is an encoder-decoder model the kwargs should include `encoder_outputs`.

    Yields
    ------
    tuple[list[str], list[dict], list[int]]
    """
    # init values
    logits_processor = (logits_processor if logits_processor is not None
                        else LogitsProcessorList())
    stopping_criteria = (stopping_criteria if stopping_criteria is not None
                         else StoppingCriteriaList())
    if max_length is not None:
        warnings.warn(
            "`max_length` is deprecated in this function, use "
            "`stopping_criteria=StoppingCriteriaList("
            "[MaxLengthCriteria(max_length=max_length)])` instead.",
            UserWarning,
        )
        stopping_criteria = validate_stopping_criteria(stopping_criteria,
                                                       max_length)
    pad_token_id = (pad_token_id if pad_token_id is not None
                    else model.config.pad_token_id)
    eos_token_id = (eos_token_id if eos_token_id is not None
                    else model.config.eos_token_id)
    output_scores = (output_scores if output_scores is not None
                     else model.config.output_scores)
    output_attentions = (output_attentions if output_attentions is not None
                         else model.config.output_attentions)
    output_hidden_states = (
        output_hidden_states if output_hidden_states
        is not None else model.config.output_hidden_states
    )
    return_dict_in_generate = (
        return_dict_in_generate if return_dict_in_generate
        is not None else model.config.return_dict_in_generate
    )

    # init attention / hidden states / scores tuples
    scores = () if (return_dict_in_generate and output_scores) else None
    decoder_attentions = () if (
                return_dict_in_generate and output_attentions) else None
    cross_attentions = () if (
                return_dict_in_generate and output_attentions) else None
    decoder_hidden_states = () if (
                return_dict_in_generate and output_hidden_states) else None

    # If model is an encoder-decoder, retrieve encoder attention weights and
    # hidden states.
    if return_dict_in_generate and model.config.is_encoder_decoder:
        encoder_attentions = model_kwargs["encoder_outputs"].get(
            "attentions") if output_attentions else None
        encoder_hidden_states = (
            model_kwargs["encoder_outputs"].get(
                "hidden_states") if output_hidden_states else None
        )

    # Keep track of which sequences are already finished.
    unfinished_sequences = input_ids.new(input_ids.shape[0]).fill_(1)
    cur_len = input_ids.shape[-1]

    this_peer_finished = False  # used by synced_gpus only
    while True:

        if synced_gpus:
            # Under synced_gpus the `forward` call must continue until all
            # gpus complete their sequence. The following logic allows an early
            # break if all peers finished generating their sequence
            this_peer_finished_flag = torch.tensor(
                0.0 if this_peer_finished else 1.0).to(input_ids.device)
            # send 0.0 if we finished, 1.0 otherwise
            dist.all_reduce(this_peer_finished_flag, op=dist.ReduceOp.SUM)
            # did all peers finish? the reduced sum will be 0.0 then
            if this_peer_finished_flag.item() == 0.0:
                break

        # prepare model inputs
        model_inputs = model.prepare_inputs_for_generation(input_ids,
                                                           **model_kwargs)

        # forward pass to get next token
        outputs = model(
            **model_inputs,
            return_dict=True,
            output_attentions=output_attentions,
            output_hidden_states=output_hidden_states,
        )

        # Don't waste resources running the code we don't need.
        if synced_gpus and this_peer_finished:
            cur_len = cur_len + 1
            continue

        next_token_logits = outputs.logits[:, -1, :]

        # Store scores, attentions and hidden_states when required.
        if return_dict_in_generate:
            if output_scores:
                scores += (next_token_logits,)
            if output_attentions:
                decoder_attentions += (
                    (outputs.decoder_attentions,)
                    if model.config.is_encoder_decoder else (
                    outputs.attentions,)
                )
                if model.config.is_encoder_decoder:
                    cross_attentions += (outputs.cross_attentions,)

            if output_hidden_states:
                decoder_hidden_states += (
                    (outputs.decoder_hidden_states,)
                    if model.config.is_encoder_decoder
                    else (outputs.hidden_states,)
                )

        # pre-process distribution
        next_tokens_scores = logits_processor(input_ids, next_token_logits)

        # argmax
        next_tokens = torch.argmax(next_tokens_scores, dim=-1)

        # finished sentences should have their next token be a padding token
        if eos_token_id is not None:
            if pad_token_id is None:
                raise ValueError(
                    "If `eos_token_id` is defined, make sure that "
                    "`pad_token_id` is defined."
                )
            next_tokens = next_tokens * unfinished_sequences + pad_token_id * (
                        1 - unfinished_sequences)

        probs = nn.functional.softmax(next_tokens_scores, dim=-1)

        # Compute items to yield now, but wait until `unfinished_sequences` is
        # updated several lines below before yielding. Huggingface sets
        # corresponding unfinished_sequences val to 0 on the token AFTER
        # EOS_TOKEN, but we want it to be 0 starting WITH that token. Don't
        # compute earlier because we want to wait for next_tokens to be updated
        # first (sets finished sequence predictions to EOS) and don't do this
        # later in case any of the _top_word_probs inputs change (don't
        # think they do, but just being careful).
        next_toks, next_w2ps =_top_word_probs(probs, next_tokens,
                                              tokenizer, probs_per_token)

        # update generated ids, model inputs, and length for next step
        input_ids = torch.cat([input_ids, next_tokens[:, None]], dim=-1)
        model_kwargs = model._update_model_kwargs_for_generation(
            outputs, model_kwargs,
            is_encoder_decoder=model.config.is_encoder_decoder
        )
        cur_len = cur_len + 1

        # if eos_token was found in one sentence, set sentence to finished
        if eos_token_id is not None:
            unfinished_sequences = unfinished_sequences.mul(
                (next_tokens != eos_token_id).long())

        yield next_toks, next_w2ps, unfinished_sequences.tolist()

        # Stop when each sentence is finished, or if we exceed the max length.
        if unfinished_sequences.max() == 0 or stopping_criteria(input_ids,
                                                                scores):
            if not synced_gpus:
                break
            else:
                this_peer_finished = True


# Note: originally tried subclassing GenerationMixin but then we'd have to
# redefine a bunch of intermediate classes OR somehow change an object
# parent class at runtime via python dark arts. Decided to just leave these
# as functions.
def live_sample(
        model,
        tokenizer,
        input_ids: torch.LongTensor,
        logits_processor: Optional[LogitsProcessorList] = None,
        stopping_criteria: Optional[StoppingCriteriaList] = None,
        logits_warper: Optional[LogitsProcessorList] = None,
        max_length: Optional[int] = None,
        pad_token_id: Optional[int] = None,
        eos_token_id: Optional[int] = None,
        output_attentions: Optional[bool] = None,
        output_hidden_states: Optional[bool] = None,
        output_scores: Optional[bool] = None,
        return_dict_in_generate: Optional[bool] = None,
        synced_gpus: Optional[bool] = False,
        probs_per_token: Optional[int] = 5,
        **model_kwargs,
) -> Union[SampleOutput, torch.LongTensor]:
    r"""
    Generates sequences of token ids for models with a language modeling head using **multinomial sampling** and
    can be used for text-decoder, text-to-text, speech-to-text, and vision-to-text models.
    Parameters:
        input_ids (`torch.LongTensor` of shape `(batch_size, sequence_length)`):
            The sequence used as a prompt for the generation.
        logits_processor (`LogitsProcessorList`, *optional*):
            An instance of [`LogitsProcessorList`]. List of instances of class derived from [`LogitsProcessor`]
            used to modify the prediction scores of the language modeling head applied at each generation step.
        stopping_criteria (`StoppingCriteriaList`, *optional*):
            An instance of [`StoppingCriteriaList`]. List of instances of class derived from [`StoppingCriteria`]
            used to tell if the generation loop should stop.
        logits_warper (`LogitsProcessorList`, *optional*):
            An instance of [`LogitsProcessorList`]. List of instances of class derived from [`LogitsWarper`] used
            to warp the prediction score distribution of the language modeling head applied before multinomial
            sampling at each generation step.
        max_length (`int`, *optional*, defaults to 20):
            **DEPRECATED**. Use `logits_processor` or `stopping_criteria` directly to cap the number of generated
            tokens. The maximum length of the sequence to be generated.
        pad_token_id (`int`, *optional*):
            The id of the *padding* token.
        eos_token_id (`int`, *optional*):
            The id of the *end-of-sequence* token.
        output_attentions (`bool`, *optional*, defaults to `False`):
            Whether or not to return the attentions tensors of all attention layers. See `attentions` under
            returned tensors for more details.
        output_hidden_states (`bool`, *optional*, defaults to `False`):
            Whether or not to return the hidden states of all layers. See `hidden_states` under returned tensors
            for more details.
        output_scores (`bool`, *optional*, defaults to `False`):
            Whether or not to return the prediction scores. See `scores` under returned tensors for more details.
        return_dict_in_generate (`bool`, *optional*, defaults to `False`):
            Whether or not to return a [`~utils.ModelOutput`] instead of a plain tuple.
        synced_gpus (`bool`, *optional*, defaults to `False`):
            Whether to continue running the while loop until max_length (needed for ZeRO stage 3)
        model_kwargs:
            Additional model specific kwargs will be forwarded to the `forward` function of the model. If model is
            an encoder-decoder model the kwargs should include `encoder_outputs`.

    Yields
    ------
    tuple[list[str], list[dict], list[int]]
    """
    # init values
    logits_processor = (logits_processor if logits_processor is not None
                        else LogitsProcessorList())
    stopping_criteria = (stopping_criteria if stopping_criteria is not None
                         else StoppingCriteriaList())
    if max_length is not None:
        warnings.warn(
            "`max_length` is deprecated in this function, use"
            " `stopping_criteria=StoppingCriteriaList("
            "MaxLengthCriteria(max_length=max_length))` instead.",
            UserWarning,
        )
        stopping_criteria = validate_stopping_criteria(stopping_criteria,
                                                       max_length)
    logits_warper = (logits_warper if logits_warper is not None
                     else LogitsProcessorList())
    pad_token_id = (pad_token_id if pad_token_id is not None
                    else model.config.pad_token_id)
    eos_token_id = (eos_token_id if eos_token_id is not None
                    else model.config.eos_token_id)
    output_scores = (output_scores if output_scores is not None
                     else model.config.output_scores)
    output_attentions = (output_attentions if output_attentions is not None
                         else model.config.output_attentions)
    output_hidden_states = (
        output_hidden_states if output_hidden_states is not None
        else model.config.output_hidden_states
    )
    return_dict_in_generate = (
        return_dict_in_generate if return_dict_in_generate is not None
        else model.config.return_dict_in_generate
    )

    # init attention / hidden states / scores tuples
    scores = () if (return_dict_in_generate and output_scores) else None
    decoder_attentions = () if (
                return_dict_in_generate and output_attentions) else None
    cross_attentions = () if (
                return_dict_in_generate and output_attentions) else None
    decoder_hidden_states = () if (
                return_dict_in_generate and output_hidden_states) else None

    # If model is an encoder-decoder, retrieve encoder attention weights and
    # hidden states.
    if return_dict_in_generate and model.config.is_encoder_decoder:
        encoder_attentions = model_kwargs["encoder_outputs"].get(
            "attentions") if output_attentions else None
        encoder_hidden_states = (
            model_kwargs["encoder_outputs"].get(
                "hidden_states") if output_hidden_states else None
        )

    # keep track of which sequences are already finished
    unfinished_sequences = input_ids.new(input_ids.shape[0]).fill_(1)

    this_peer_finished = False  # used by synced_gpus only
    # auto-regressive generation
    while True:
        if synced_gpus:
            # Under synced_gpus the `forward` call must continue until all
            # gpus complete their sequence. The following logic allows an early
            # break if all peers finished generating their sequence
            this_peer_finished_flag = torch.tensor(
                0.0 if this_peer_finished else 1.0).to(input_ids.device)
            # send 0.0 if we finished, 1.0 otherwise
            dist.all_reduce(this_peer_finished_flag, op=dist.ReduceOp.SUM)
            # did all peers finish? the reduced sum will be 0.0 then
            if this_peer_finished_flag.item() == 0.0:
                break

        # prepare model inputs
        model_inputs = model.prepare_inputs_for_generation(input_ids,
                                                           **model_kwargs)

        # forward pass to get next token
        outputs = model(
            **model_inputs,
            return_dict=True,
            output_attentions=output_attentions,
            output_hidden_states=output_hidden_states,
        )

        if synced_gpus and this_peer_finished:
            continue  # don't waste resources running the code we don't need

        next_token_logits = outputs.logits[:, -1, :]

        # pre-process distribution
        next_token_scores = logits_processor(input_ids, next_token_logits)
        next_token_scores = logits_warper(input_ids, next_token_scores)

        # Store scores, attentions and hidden_states when required
        if return_dict_in_generate:
            if output_scores:
                scores += (next_token_scores,)
            if output_attentions:
                decoder_attentions += (
                    (outputs.decoder_attentions,)
                    if model.config.is_encoder_decoder else
                    (outputs.attentions,)
                )
                if model.config.is_encoder_decoder:
                    cross_attentions += (outputs.cross_attentions,)

            if output_hidden_states:
                decoder_hidden_states += (
                    (outputs.decoder_hidden_states,)
                    if model.config.is_encoder_decoder
                    else (outputs.hidden_states,)
                )

        # sample
        # Shape: (num_return_sequences, vocab_size)
        probs = nn.functional.softmax(next_token_scores, dim=-1)
        # Shape: (num_return_sequences,)
        next_tokens = torch.multinomial(probs, num_samples=1).squeeze(1)

        # finished sentences should have their next token be a padding token
        if eos_token_id is not None:
            if pad_token_id is None:
                raise ValueError(
                    "If `eos_token_id` is defined, make sure that "
                    "`pad_token_id` is defined."
                )
            next_tokens = next_tokens * unfinished_sequences + pad_token_id * (
                        1 - unfinished_sequences)

        # Compute items to yield now, but wait until `unfinished_sequences` is
        # updated several lines below before yielding. Huggingface sets
        # corresponding unfinished_sequences val to 0 on the token AFTER
        # EOS_TOKEN, but we want it to be 0 starting WITH that token. Don't
        # compute earlier because we want to wait for next_tokens to be updated
        # first (sets finished sequence predictions to EOS) and don't do this
        # later in case any of the _top_word_probs inputs change (don't
        # think they do, but just being careful).
        next_toks, next_w2ps =_top_word_probs(probs, next_tokens,
                                              tokenizer, probs_per_token)

        # update generated ids, model inputs, and length for next step
        input_ids = torch.cat([input_ids, next_tokens[:, None]], dim=-1)
        model_kwargs = model._update_model_kwargs_for_generation(
            outputs, model_kwargs,
            is_encoder_decoder=model.config.is_encoder_decoder
        )

        # if eos_token was found in one sentence, set sentence to finished
        if eos_token_id is not None:
            unfinished_sequences = unfinished_sequences.mul(
                (next_tokens != eos_token_id).long())

        # Shape: (num_return_sequences, probs_per_word)
        yield next_toks, next_w2ps, unfinished_sequences.tolist()

        # Stop when each sentence is finished, or if we exceed the max length.
        if unfinished_sequences.max() == 0 or stopping_criteria(input_ids,
                                                                scores):
            if not synced_gpus:
                break
            else:
                this_peer_finished = True


@torch.no_grad()
def live_generate(
        model,
        tokenizer,
        inputs: Optional[torch.Tensor] = None,
        max_length: Optional[int] = None,
        min_length: Optional[int] = None,
        do_sample: Optional[bool] = None,
        temperature: Optional[float] = None,
        top_k: Optional[int] = None,
        top_p: Optional[float] = None,
        typical_p: Optional[float] = None,
        repetition_penalty: Optional[float] = None,
        bad_words_ids: Optional[Iterable[int]] = None,
        bos_token_id: Optional[int] = None,
        pad_token_id: Optional[int] = None,
        eos_token_id: Optional[int] = None,
        no_repeat_ngram_size: Optional[int] = None,
        encoder_no_repeat_ngram_size: Optional[int] = None,
        num_return_sequences: Optional[int] = None,
        max_time: Optional[float] = None,
        max_new_tokens: Optional[int] = None,
        decoder_start_token_id: Optional[int] = None,
        use_cache: Optional[bool] = None,
        diversity_penalty: Optional[float] = None,
        logits_processor: Optional[LogitsProcessorList] = None,
        renormalize_logits: Optional[bool] = None,
        stopping_criteria: Optional[StoppingCriteriaList] = None,
        output_attentions: Optional[bool] = None,
        output_hidden_states: Optional[bool] = None,
        output_scores: Optional[bool] = None,
        return_dict_in_generate: Optional[bool] = None,
        forced_bos_token_id: Optional[int] = None,
        forced_eos_token_id: Optional[int] = None,
        remove_invalid_values: Optional[bool] = None,
        synced_gpus: Optional[bool] = False,
        exponential_decay_length_penalty: Optional[Tuple[int, float]] = None,
        suppress_tokens: Optional[List[int]] = None,
        begin_suppress_tokens: Optional[List[int]] = None,
        forced_decoder_ids: Optional[List[List[int]]] = None,
        probs_per_token: Optional[int] = 5,
        **model_kwargs,
) -> Union[GenerateOutput, torch.LongTensor]:
    r"""
    Generates sequences of token ids for models with a language modeling head. The method supports the following
    generation methods for text-decoder, text-to-text, speech-to-text, and vision-to-text models:
        - *greedy decoding* by calling [`~generation_utils.GenerationMixin.greedy_search`]
            if `do_sample=False`.
        - *multinomial sampling* by calling [`~generation_utils.GenerationMixin.sample`]
            if `do_sample=True`.
    <Tip warning={true}>
    Apart from `inputs`, all the arguments below will default to the value of the attribute of the same name as
    defined in the model's config (`config.json`) which in turn defaults to the
    [`~modeling_utils.PretrainedConfig`] of the model.
    </Tip>
    Most of these parameters are explained in more detail in [this blog
    post](https://huggingface.co/blog/how-to-generate).
    Parameters:
        inputs (`torch.Tensor` of varying shape depending on the modality,
                *optional*. HDM update: added support for passing in str. Only
                1 prompt can be provided at a time (at least currently).):
            The sequence used as a prompt for the generation or as model inputs to the encoder. If `None` the
            method initializes it with `bos_token_id` and a batch size of 1. For decoder-only models `inputs`
            should of in the format of `input_ids`. For encoder-decoder models *inputs* can represent any of
            `input_ids`, `input_values`, `input_features`, or `pixel_values`.
        max_length (`int`, *optional*, defaults to `model.config.max_length`):
            The maximum length the generated tokens can have. Corresponds to the length of the input prompt +
            `max_new_tokens`. In general, prefer the use of `max_new_tokens`, which ignores the number of tokens in
            the prompt.
        max_new_tokens (`int`, *optional*):
            The maximum numbers of tokens to generate, ignoring the number of tokens in the prompt.
        min_length (`int`, *optional*, defaults to `model.config.min_length` or 10 if the config does not set any value):
            The minimum length of the sequence to be generated.
        do_sample (`bool`, *optional*, defaults to `model.config.do_sample` or `False` if the config does not set any value):
            Whether or not to use sampling ; use greedy decoding otherwise.
        temperature (`float`, *optional*, defaults to `model.config.temperature` or 1.0 if the config does not set any value):
            The value used to module the next token probabilities.
        penalty_alpha (`float`, *optional*, defaults to `model.config.penalty_alpha` or None if the config does not set any value):
            The values balance the model confidence and the degeneration penalty in contrastive search decoding.
        top_k (`int`, *optional*, defaults to `model.config.top_k` or 50 if the config does not set any value):
            The number of highest probability vocabulary tokens to keep for top-k-filtering.
        top_p (`float`, *optional*, defaults to `model.config.top_p` or 1.0 if the config does not set any value):
            If set to float < 1, only the smallest set of most probable tokens with probabilities that add up to
            `top_p` or higher are kept for generation.
        typical_p (`float`, *optional*, defaults to `model.config.typical_p` or 1.0 if the config does not set any value):
            The amount of probability mass from the original distribution to be considered in typical decoding. If
            set to 1.0 it takes no effect. See [this paper](https://arxiv.org/pdf/2202.00666.pdf) for more details.
        repetition_penalty (`float`, *optional*, defaults to `model.config.repetition_penalty` or 1.0 if the config does not set any value):
            The parameter for repetition penalty. 1.0 means no penalty. See [this
            paper](https://arxiv.org/pdf/1909.05858.pdf) for more details.
        pad_token_id (`int`, *optional*, defaults to `model.config.pad_token_id`):
            The id of the *padding* token.
        bos_token_id (`int`, *optional*, defaults to `model.config.bos_token_id`):
            The id of the *beginning-of-sequence* token.
        eos_token_id (`int`, *optional*, defaults to `model.config.eos_token_id`):
            The id of the *end-of-sequence* token.
        no_repeat_ngram_size (`int`, *optional*, defaults to `model.config.no_repeat_ngram_size` or 0 if the config does not set any value):
            If set to int > 0, all ngrams of that size can only occur once.
        encoder_no_repeat_ngram_size (`int`, *optional*, defaults to `model.config.encoder_no_repeat_ngram_size` or 0 if the config does not set any value):
            If set to int > 0, all ngrams of that size that occur in the `encoder_input_ids` cannot occur in the
            `decoder_input_ids`.
        bad_words_ids(`List[List[int]]`, *optional*, defaults to `model.config.bad_words_ids`):
            List of token ids that are not allowed to be generated. In order to get the token ids of the words that
            should not appear in the generated text, use `tokenizer(bad_words, add_prefix_space=True,
            add_special_tokens=False).input_ids`.
        num_return_sequences(`int`, *optional*, defaults to `model.config.num_return_sequences` or 1 if the config does not set any value):
            The number of independently computed returned sequences for each element in the batch.
        max_time(`float`, *optional*):
            The maximum amount of time you allow the computation to run for in seconds. generation will still
            finish the current pass after allocated time has been passed.
        attention_mask (`torch.LongTensor` of shape `(batch_size, sequence_length)`, *optional*):
            Mask to avoid performing attention on padding token indices. Mask values are in `[0, 1]`, 1 for tokens
            that are not masked, and 0 for masked tokens. If not provided, will default to a tensor the same shape
            as `input_ids` that masks the pad token. [What are attention masks?](../glossary#attention-mask)
        decoder_start_token_id (`int`, *optional*):
            If an encoder-decoder model starts decoding with a different token than *bos*, the id of that token.
        use_cache (`bool`, *optional*, defaults to `True`):
            Whether or not the model should use the past last key/values attentions (if applicable to the model) to
            speed up decoding.
        logits_processor (`LogitsProcessorList`, *optional*):
             Custom logits processors that complement the default logits processors built from arguments and a
             model's config. If a logit processor is passed that is already created with the arguments or a model's
             config an error is thrown. This feature is intended for advanced users.
        renormalize_logits (`bool`, *optional*, defaults to `False`):
            Whether to renormalize the logits after applying all the logits processors or warpers (including the
            custom ones). It's highly recommended to set this flag to `True` as the search algorithms suppose the
            score logits are normalized but some logit processors or warpers break the normalization.
        stopping_criteria (`StoppingCriteriaList`, *optional*):
             Custom stopping criteria that complement the default stopping criteria built from arguments and a
             model's config. If a stopping criteria is passed that is already created with the arguments or a
             model's config an error is thrown. This feature is intended for advanced users.
        output_attentions (`bool`, *optional*, defaults to `model.config.output_attentions` or `False` if the config does not set any value):
            Whether or not to return the attentions tensors of all attention layers. See `attentions` under
            returned tensors for more details.
        output_hidden_states (`bool`, *optional*, defaults to `model.config.output_hidden_states` or `False` if the config does not set any value):
            Whether or not to return the hidden states of all layers. See `hidden_states` under returned tensors
            for more details.
        output_scores (`bool`, *optional*, defaults to `model.config.output_scores` or `False` if the config does not set any value):
            Whether or not to return the prediction scores. See `scores` under returned tensors for more details.
        return_dict_in_generate (`bool`, *optional*, defaults to `model.config.return_dict_in_generate` or `False` if the config does not set any value):
            Whether or not to return a [`~utils.ModelOutput`] instead of a plain tuple.
        forced_bos_token_id (`int`, *optional*, defaults to `model.config.forced_bos_token_id`):
            The id of the token to force as the first generated token after the `decoder_start_token_id`. Useful
            for multilingual models like [mBART](../model_doc/mbart) where the first generated token needs to be
            the target language token.
        forced_eos_token_id (`int`, *optional*, defaults to `model.config.forced_eos_token_id`):
            The id of the token to force as the last generated token when `max_length` is reached.
        remove_invalid_values (`bool`, *optional*, defaults to `model.config.remove_invalid_values`):
            Whether to remove possible *nan* and *inf* outputs of the model to prevent the generation method to
            crash. Note that using `remove_invalid_values` can slow down generation.
        synced_gpus (`bool`, *optional*, defaults to `False`):
            Whether to continue running the while loop until max_length (needed for ZeRO stage 3)
        exponential_decay_length_penalty (`tuple(int, float)`, *optional*, defaults to `model.config.exponential_decay_length_penalty`):
            This Tuple adds an exponentially increasing length penalty, after a certain amount of tokens have been
            generated. The tuple shall consist of: `(start_index, decay_factor)` where `start_index` indicates
            where penalty starts and `decay_factor` represents the factor of exponential decay
        suppress_tokens  (`List[int]`, *optional*, defaults to `model.config.suppress_tokens`):
            A list of tokens that will be supressed at generation. The `SupressTokens` logit processor will set
            their log probs to `-inf` so that they are not sampled.
        begin_suppress_tokens  (`List[int]`, *optional*, defaults to `model.config.begin_suppress_tokens`):
            A list of tokens that will be supressed at the begining of the generation. The `SupressBeginTokens`
            logit processor will set their log probs to `-inf` so that they are not sampled.
        forced_decoder_ids (`List[List[int]]`, *optional*, defaults to `model.config.forced_decoder_ids`):
            A list of pairs of integers which indicates a mapping from generation indices to token indices that
            will be forced before sampling. For example, `[[1, 123]]` means the second generated token will always
            be a token of index 123.
        model_kwargs:
            Additional model specific kwargs will be forwarded to the `forward` function of the model. If the model
            is an encoder-decoder model, encoder specific kwargs should not be prefixed and decoder specific kwargs
            should be prefixed with *decoder_*.

    Yields
    ------
    tuple[list[str], list[dict], list[int]]: Each item has num_return_sequences
    items. First list contains predictions for next token(s). Second item
    contains word->prob dict for each next token (top k words + the sampled
    word). Third item shows which sequences are alive: 1 means we haven't hit
    EOS yet, 0 means we have.

    Examples:

    Multinomial Sampling:
    ```python
    >>> from transformers import AutoTokenizer, AutoModelForCausalLM
    >>> from transformers import AutoTokenizer, AutoModelForCausalLM
    >>> import torch
    >>> tokenizer = AutoTokenizer.from_pretrained("gpt2")
    >>> model = AutoModelForCausalLM.from_pretrained("gpt2")
    >>> prompt = "Today I believe we can finally"
    >>> input_ids = tokenizer(prompt, return_tensors="pt").input_ids
    >>> for next_token, tok2prob in live_generate(
            model, tokenizer, input_ids, do_sample=True, max_length=30)
    >>> ):
    >>>     print(next_token)
    ```
    """
    unsupported = (
        'num_beams',
        'num_beam_groups',
        'prefix_allowed_tokens_fn',
    )
    for name in unsupported:
        if name in model_kwargs:
            raise ValueError(f'You passed in {name} but that '
                             'argument is not supported.')
    num_beams = 1
    num_beam_groups = 1
    prefix_allowed_tokens_fn = None

    if isinstance(inputs, str):
        inputs = tokenizer(inputs, return_tensors='pt').input_ids
    if isinstance(inputs, list) or inputs.shape[0] > 1:
        raise NotImplementedError('We currently only support live generation '
                                  'for one input prompt at a time.')
    inputs = inputs.to(model.device)

    # 1. Set generation parameters if not already defined
    bos_token_id = (bos_token_id if bos_token_id is not None
                    else model.config.bos_token_id)

    do_sample = do_sample if do_sample is not None else model.config.do_sample
    num_return_sequences = (
        num_return_sequences if num_return_sequences is not None
        else model.config.num_return_sequences
    )
    logits_processor = (logits_processor if logits_processor is not None
                        else LogitsProcessorList())
    stopping_criteria = (stopping_criteria if stopping_criteria is not None
                         else StoppingCriteriaList())

    pad_token_id = (pad_token_id if pad_token_id is not None
                    else model.config.pad_token_id)
    eos_token_id = (eos_token_id if eos_token_id is not None
                    else model.config.eos_token_id)

    if eos_token_id is None and hasattr(model.config, "decoder"):
        eos_token_id = model.config.decoder.eos_token_id

    if pad_token_id is None and eos_token_id is not None:
        if model_kwargs.get("attention_mask", None) is None:
            logger.warning(
                "The attention mask and the pad token id were not set. "
                "As a consequence, you may observe "
                "unexpected behavior. Please pass your input's "
                "`attention_mask` to obtain reliable results."
            )
        logger.warning(
            f"Setting `pad_token_id` to `eos_token_id`:"
            f"{eos_token_id} for open-end generation."
        )
        pad_token_id = eos_token_id

    output_scores = (output_scores if output_scores is not None
                     else model.config.output_scores)
    output_attentions = (output_attentions if output_attentions is not None
                         else model.config.output_attentions)
    output_hidden_states = (
        output_hidden_states if output_hidden_states is not None
        else model.config.output_hidden_states
    )
    return_dict_in_generate = (
        return_dict_in_generate if return_dict_in_generate is not None
        else model.config.return_dict_in_generate
    )

    # 2. Define model inputs
    # inputs_tensor has to be defined
    # model_input_name is defined if model-specific keyword input is passed
    # otherwise model_input_name is None
    # all model-specific keyword inputs are removed from `model_kwargs`
    inputs_tensor, model_input_name, model_kwargs = model._prepare_model_inputs(
        inputs, bos_token_id, model_kwargs
    )
    batch_size = inputs_tensor.shape[0]

    # 3. Define other model kwargs
    model_kwargs["output_attentions"] = output_attentions
    model_kwargs["output_hidden_states"] = output_hidden_states
    model_kwargs["use_cache"] = use_cache

    accepts_attention_mask = "attention_mask" in set(
        inspect.signature(model.forward).parameters.keys())
    requires_attention_mask = "encoder_outputs" not in model_kwargs

    if (model_kwargs.get("attention_mask", None) is None
            and requires_attention_mask and accepts_attention_mask):
        model_kwargs[
            "attention_mask"] = model._prepare_attention_mask_for_generation(
            inputs_tensor, pad_token_id, eos_token_id
        )

    # decoder-only models should use left-padding for generation
    if not model.config.is_encoder_decoder:
        if pad_token_id is not None and torch.sum(
                inputs_tensor[:, -1] == pad_token_id) > 0:
            logger.warning(
                "A decoder-only architecture is being used, but right-padding "
                "was detected! For correct generation results, please set "
                "`padding_side='left'` when initializing the tokenizer."
            )

    if model.config.is_encoder_decoder and "encoder_outputs" not in model_kwargs:
        # if model is encoder decoder encoder_outputs are created
        # and added to `model_kwargs`
        model_kwargs = model._prepare_encoder_decoder_kwargs_for_generation(
            inputs_tensor, model_kwargs, model_input_name
        )

    # 4. Prepare `input_ids` which will be used for auto-regressive generation
    if model.config.is_encoder_decoder:
        input_ids = model._prepare_decoder_input_ids_for_generation(
            batch_size,
            decoder_start_token_id=decoder_start_token_id,
            bos_token_id=bos_token_id,
            model_kwargs=model_kwargs,
            device=inputs_tensor.device,
        )
    else:
        # if decoder-only then inputs_tensor has to be `input_ids`
        input_ids = inputs_tensor

    # 5. Prepare `max_length` depending on other stopping criteria.
    input_ids_seq_length = input_ids.shape[-1]
    if max_length is None and max_new_tokens is None:
        warnings.warn(
            "Neither `max_length` nor `max_new_tokens` has been set, "
            "`max_length` will default to " f"{model.config.max_length} "
            f"(`model.config.max_length`). Controlling `max_length` via the "
            "config is deprecated and `max_length` will be removed from "
            "the config in v5 of Transformers -- we recommend "
            "using `max_new_tokens` to control the maximum length "
            "of the generation.",
            UserWarning,
        )
    elif max_length is None and max_new_tokens is not None:
        max_length = max_new_tokens + input_ids_seq_length
    elif max_length is not None and max_new_tokens is not None:
        raise ValueError(
            "Both `max_new_tokens` and `max_length` have been set but they "
            "serve the same purpose -- setting a"
            " limit to the generated output length. Remove one of those "
            "arguments. Please refer to thedocumentation for more "
            " information. "
            "(https://huggingface.co/docs/transformers/main/en/main_classes/text_generation)"
        )
    # default to config if still None
    max_length = max_length if max_length is not None else model.config.max_length
    min_length = min_length if min_length is not None else model.config.min_length

    if min_length is not None and min_length > max_length:
        raise ValueError(
            f"Unfeasible length constraints: the minimum length "
            f"({min_length}) is larger than the maximum "
            f"length ({max_length})"
        )
    if input_ids_seq_length >= max_length:
        input_ids_string = ("decoder_input_ids" if
                            model.config.is_encoder_decoder else "input_ids")
        logger.warning(
            f"Input length of {input_ids_string} is {input_ids_seq_length}, "
            f"but `max_length` is set to"
            f" {max_length}. This can lead to unexpected behavior. "
            f"You should consider increasing `max_new_tokens`."
        )

    # 6. determine generation mode
    is_greedy_gen_mode = not do_sample
    if model.device.type != input_ids.device.type:
        warnings.warn(
            "You are calling .generate() with the `input_ids` being on a "
            "device type different"
            f" than your model's device. `input_ids` is on "
            f"{input_ids.device.type}, whereas the model"
            f" is on {model.device.type}. You may experience unexpected "
            f"behaviors or slower generation."
            " Please make sure that you have put `input_ids` to the"
            f" correct device by calling for example "
            f"input_ids = input_ids.to('{model.device.type}') before"
            " running `live_generate()`.",
            UserWarning,
        )

    # 7. prepare distribution pre_processing samplers
    logits_processor = model._get_logits_processor(
        repetition_penalty=repetition_penalty,
        no_repeat_ngram_size=no_repeat_ngram_size,
        encoder_no_repeat_ngram_size=encoder_no_repeat_ngram_size,
        input_ids_seq_length=input_ids_seq_length,
        encoder_input_ids=inputs_tensor,
        bad_words_ids=bad_words_ids,
        min_length=min_length,
        max_length=max_length,
        eos_token_id=eos_token_id,
        forced_bos_token_id=forced_bos_token_id,
        forced_eos_token_id=forced_eos_token_id,
        prefix_allowed_tokens_fn=prefix_allowed_tokens_fn,
        num_beams=num_beams,
        num_beam_groups=num_beam_groups,
        diversity_penalty=diversity_penalty,
        remove_invalid_values=remove_invalid_values,
        exponential_decay_length_penalty=exponential_decay_length_penalty,
        logits_processor=logits_processor,
        renormalize_logits=renormalize_logits,
    )

    # 8. prepare stopping criteria
    stopping_criteria = model._get_stopping_criteria(
        max_length=max_length, max_time=max_time,
        stopping_criteria=stopping_criteria
    )
    # 9. go into different generation modes
    if is_greedy_gen_mode:
        if num_return_sequences > 1:
            raise ValueError(
                f"num_return_sequences has to be 1, but is "
                f"{num_return_sequences} when doing greedy search."
            )

        # 10. run greedy search
        yield from live_greedy_search(
            model,
            tokenizer,
            input_ids,
            logits_processor=logits_processor,
            stopping_criteria=stopping_criteria,
            pad_token_id=pad_token_id,
            eos_token_id=eos_token_id,
            output_scores=output_scores,
            return_dict_in_generate=return_dict_in_generate,
            synced_gpus=synced_gpus,
            **model_kwargs,
        )

    else:
        # 10. prepare logits warper
        logits_warper = model._get_logits_warper(
            top_k=top_k,
            top_p=top_p,
            typical_p=typical_p,
            temperature=temperature,
            num_beams=num_beams,
            renormalize_logits=renormalize_logits,
        )

        # 11. expand input_ids with `num_return_sequences` additional
        # sequences per batch
        input_ids, model_kwargs = model._expand_inputs_for_generation(
            input_ids=input_ids,
            expand_size=num_return_sequences,
            is_encoder_decoder=model.config.is_encoder_decoder,
            **model_kwargs,
        )

        # 12. run sample
        yield from live_sample(
            model,
            tokenizer,
            input_ids,
            logits_processor=logits_processor,
            logits_warper=logits_warper,
            stopping_criteria=stopping_criteria,
            pad_token_id=pad_token_id,
            eos_token_id=eos_token_id,
            output_scores=output_scores,
            return_dict_in_generate=return_dict_in_generate,
            synced_gpus=synced_gpus,
            **model_kwargs,
        )


def truncate_if_necessary(tokenizer, text, max_new_tokens,
                          do_strip=True):
    """Truncate text prompt so that the requested generation length can
    execute without error (this is limited by model_max_length).

    Parameters
    ----------
    tokenizer
    text: str
    max_new_tokens: int
    do_strip: bool

    Returns
    -------
    dict
    """
    if do_strip:
        text = text.rstrip()
    xb = tokenizer(text, return_tensors='pt')
    ids = xb.input_ids.squeeze(0)
    original_len = len(ids)
    original_len_words = len(text.split())
    ids = ids[-tokenizer.model_max_length + max_new_tokens:].squeeze(0)
    text = tokenizer.decode(ids)
    new_len = len(ids)
    new_len_words = len(text.split())
    res = {
        'text': text,
        # Add batch dimension back in so we can more easily pass it to
        # live_generate.
        'input_ids': ids.unsqueeze(0),
        'original_length': original_len,
        'original_length_words': original_len_words,
        'truncated_length': new_len,
        'truncated_length_words': new_len_words,
    }
    return res


def word_to_ids(tokenizer, word):
    """Convert a word (str) to a list of word indices (ints). Using this in
    `generate` lets user pass in list of strings for words to skip/force
    instead of list[list[int]].

    Parameters
    ----------
    tokenizer: transformers Tokenizer
    word: str

    Returns
    -------
    list[int]
    """
    return tokenizer.convert_tokens_to_ids(
        tokenizer.tokenize(word)
    )


@add_docstring(live_generate)
def generate(
        model,
        tokenizer,
        prompt,
        out_dir=None,
        mode_pre='a',
        max_new_tokens=50,
        repetition_penalty=1.2,
        early_stopping=False,
        do_sample=True,
        temperature=0.7,
        top_p=.99,
        **kwargs
):
    """
    Parameters
    ----------
    model
    tokenizer
    prompt: str
        Notice that you must pass in a single prompt (str), not a list of
        prompts.
    out_dir: str or Path or None
        If not None, a file will be saved in this directory with the generated
        output(s).
    mode_pre: str
        'a' or 'w'. If out_dir is provided, this determines what mode to use
        when writing to that file (append or write). Append is the default
        because the immediate use case is to use this in a callback that will
        be executing repeatedly during training.
    repetition_penalty
    early_stopping
    do_sample
    temperature
    top_p
    kwargs: any
        Additional kwargs for model.generate().

    Returns
    -------
    list[dict]: Each dict contains data for one generated sequence (we can use
    `num_return_sequences` arg to request > 1). Each dict has keys:
        tokens: list[str]
            List of generated tokens.
        probs: list[dict[str, float]]
            One dict for each time step in generation. Each dict maps token to
            its corresponding probability. Contains 5 most likely tokens and
            the selected token (when not in greedy mode, this means you can get
            either 5 or 6 probs in each dict).
        text: str
            The de-tokenized generated sequence as a single str.
        full_text: str
            Same as text but with input prompt too. Prompt and generation
            are distinguished by sub-headers PROMPT and RESPONSE. Useful when
            saving output to a file (in fact, when out_dir is provided, this
            is basically what we save, with the addition of some spacers and
            numbering).
    """
    if not isinstance(prompt, str):
        raise TypeError(f'Expected prompt to be str, got {type(prompt)}.')
    if 'inputs' in kwargs:
        del kwargs['inputs']
        warnings.warn('You should provide a str/list `prompt` instead of the '
                      'huggingface `inputs` arg.')

    n = kwargs.get('num_return_sequences', 1)
    kwargs.update(
        repetition_penalty=repetition_penalty,
        early_stopping=early_stopping,
        do_sample=do_sample,
        temperature=temperature,
        top_p=top_p,
        max_new_tokens=max_new_tokens
    )

    # Make it easier to provide words to skip/include. Lets us pass in lists
    # of strings instead of lists of lists of token IDs.
    # HF does not accept empty list, has to be None if falsy.
    kwargs['bad_words_ids'] = [word_to_ids(tokenizer, t) for t in
                               kwargs.pop('skip_words', [])] or None
    kwargs['force_words_ids'] = [word_to_ids(tokenizer, t) for t in
                                 kwargs.pop('force_words', [])] or None
    if kwargs['force_words_ids']:
        if kwargs.get('do_sample', True):
            warnings.warn('Auto-setting do_sample=False because force_words '
                          'were passed in.')
            kwargs['do_sample'] = False
        if kwargs.get('num_beams', 1) <= 1:
            warnings.warn('Auto-setting num_beams=2 because force_words were '
                          'passed in.')
            kwargs['num_beams'] = 2

    # res[i] is a dict of results for the i'th generated sequence.
    res = [{k: [] for k in ('tokens', 'probs')} for _ in range(n)]
    for word, probs, is_alive in live_generate(model, tokenizer, prompt,
                                               **kwargs):
        for cur_word, cur_probs, cur_alive, cur_res in zip(word, probs,
                                                           is_alive, res):
            if cur_alive:
                cur_res['tokens'].append(cur_word)
                cur_res['probs'].append(cur_probs)
    for i in range(len(res)):
        res[i]['text'] = tokenizer.convert_tokens_to_string(res[i]['tokens'])
        res[i]['full_text'] = ('# PROMPT\n' + prompt + '\n\n# RESPONSE\n'
                               + res[i]['text'])
    if out_dir:
        out_dir = Path(out_dir)
        save(
            ''.join(f"{i}.\n{row['full_text']}{spacer()}\n"
                    for i, row in enumerate(res, 1)),
            out_dir/'generated_text_samples.txt',
            mode_pre=mode_pre
        )
    return res


"""Functionality to help us work with masked language models.
"""

from collections import Mapping
from itertools import product
import numpy as np
import torch
import torch.nn.functional as F
from tqdm.auto import tqdm
import warnings

from htools import eprint, timer, block_timer, bound_args, params, identity, \
    Trie, deprecated


def fill_multi_mask_auto(text, pipe, top_k=5, **kwargs):
    """Fill multiple masked tokens in an input string in an independent manner
    using the builtin Huggingface pipeline functionality. This is just a
    wrapper to reshape outputs a bit to be consistent with the interface our
    other mask filling functions follow.

    Parameters
    ----------
    text: str
        Input text that already has multiple words masked. Careful: not all
        models use the same mask token, so make sure you're using the right
        one for your pipeline.
    pipe: transformers.FillMaskPipeline
        E.g. pipe = pipeline('fill-mask', model='bert-base-cased')
    top_k: int
    kwargs: any
        Ignored. Just here to maintain a consistent interface for multi mask
        functions.

    Returns
    -------
    dict:
    'token_probs'
        a list of dicts where the i'th item in the list corresponds to the
        i'th masked token. Each dict maps the top_k predicted token strings to
        probabilities.
    'seq2probs'
        a dict mapping tuples of predicted tokens (each with length m, where m
        is the number of masked tokens in the input) to a list of m
        probabilities.
    'seq2prob'
        a dict like seq2probs, but values are the product of each token's
        predicted probabilities. We also sort this dict so the most likely
        seqs come first.
    """
    res = pipe(text, top_k=top_k)
    if isinstance(res[0], Mapping):
        res = [res]
    token_probs = [{d['token_str']: d['score'] for d in row}
                   for row in res]
    seq2probs = {row: [token_probs[i][tok] for i, tok in enumerate(row)]
                 for row in product(*token_probs)}
    return {
        'token_probs': token_probs,
        'seq2probs': seq2probs,
        'seq2prob': dict(sorted(
            [(k, np.product(v)) for k, v in seq2probs.items()],
            key=lambda x: x[1], reverse=True
        ))
    }


@deprecated(msg='Soft deprecation: fill_multi_mask_left_to_right hasn\'t been '
            'tested on mask-filling pipelines like "roberta-base", and may '
            'not work as expected due to their special space characters. We '
            'suggest using fill_multi_mask_auto or '
            'fill_multi_mask_exhaustive.')
def fill_multi_mask_left_to_right(text, pipe, top_k=5, **kwargs):
    """Fill multiple masked tokens in an input string in a dependent manner.
    Huggingface pipeline can fill multiple tokens in a single pipeline call
    but (depending on the model) this will often make those predictions
    independently.

    Fill 1 masked token at a time, starting with the left-most masked token and
    working rightwards. At each iteration, we predict on a string where the
    *previous* masked token is filled with the most likely prediction from
    the previous step. This makes the token predictions dependent rather
    than independent (hopefully leading to more context-aware completions).
    It is greedy, however - we only keep the 1 most likely token at each step.

    We make a total of m forward passes where m is the total number of
    masked tokens, and we end up with k potential sequences
    where k is the number of top tokens we keep per forward pass. (This number
    will often be *much* lower than our exhaustive mask filling strategy
    because we only keep the top 1 prediction at each unmasking step.
    We do still return a total of k*m different tokens with probabilities,
    but only k of those combinations form valid predictions.)

    Parameters
    ----------
    text: str
        Input text that already has multiple words masked. Careful: not all
        models use the same mask token, so make sure you're using the right
        one for your pipeline.
    pipe: transformers.FillMaskPipeline
        E.g. pipe = pipeline('fill-mask', model='bert-base-cased')
    top_k: int
    kwargs: any
        Ignored. Just here to maintain a consistent interface for multi mask
        functions.

    Returns
    -------
    dict:
    'token_probs'
        a list of dicts where the i'th item in the list corresponds to the
        i'th masked token. Each dict maps the top_k predicted token strings to
        probabilities.
    'seq2probs'
        a dict mapping tuples of predicted tokens (each with length m, where m
        is the number of masked tokens in the input) to a list of m
        probabilities.
    'seq2prob'
        a dict like seq2probs, but values are the product of each token's
        predicted probabilities. We also sort this dict so the most likely
        seqs come first.
    """
    xb = pipe.tokenizer(text, return_tensors='pt')
    mask_idx = pipe.get_masked_index(xb.input_ids)[:, -1].tolist()
    prev_pred = pipe.tokenizer.mask_token_id
    all_res = []
    n_mask = len(mask_idx)
    running_top_toks = []
    running_top_probs = []
    for i, prev_i in tqdm(zip(mask_idx, [None] + mask_idx),
                          total=n_mask, desc='unmask'):
        if prev_i is not None:
            xb['input_ids'][0, prev_i] = prev_pred
        with torch.no_grad():
            res = pipe.model(xb.input_ids)
        probs = F.softmax(res.logits[0, i], dim=-1)
        top_probas, top_idx = torch.topk(probs, top_k, dim=-1)
        pred_tokens = pipe.tokenizer.convert_ids_to_tokens(top_idx)
        running_top_toks.append(pred_tokens[0])
        running_top_probs.append(top_probas[0].item())
        res = dict(zip(pred_tokens, top_probas.tolist()))
        prev_pred = top_idx[0]
        all_res.append(res)

    running_top_toks.pop(-1)
    running_top_probs.pop(-1)
    total_prob = np.product(running_top_probs)
    seq2probs = {}
    seq2prob = {}
    for k, v in all_res[-1].items():
        key = tuple(running_top_toks + [k])
        seq2probs[key] = running_top_probs + [v]
        seq2prob[key] = total_prob * v
    return {
        'token_probs': all_res,
        'seq2probs': seq2probs,
        'seq2prob': seq2prob
    }


def fill_multi_mask_exhaustive(text, pipe, top_k=5, space_char=' '):
    """Given a piece of text with multiple masked tokens, unmask them from left
    to right, exploring all possible branches of the tree of top predictions.
    We need to make k^(i-1) forward passes at each step (where k is the number
    of top predicted tokens we keep from each forward pass, and i is the
    1-indexed number marking which masked token we're filling), each operating
    on a different input string. We end up with k^i possible sequences and
    \sum^{m}_{i=1} m * k^{(i - 1)}
    total forward passes.

    Parameters
    ----------
    text: str
        Input text that already has multiple words masked. Careful: not all
        models use the same mask token, so make sure you're using the right
        one for your pipeline.
    pipe: transformers.FillMaskPipeline
        E.g. pipe = pipeline('fill-mask', model='bert-base-cased')
    top_k: int
    space_char: str
        Character tokenizer uses to represent space. For bert etc. this should
        be a string containing a single space. For roberta etc this should
        be 'Ġ'.

    Returns
    -------
    dict:
    'token_probs'
        a list of lists of dicts, where the i'th nested list maps to the
        i'th masked token. Dict j within a nested list maps token strings to
        predicted probabilities for the j'th pipeline call for that mask index.
        E.g. res['token_probs'][2][0] contains token probs for the 3rd masked
        token, if we filled the 0th and 1st mask spots with their single most
        likely predictions. NOTE: this is different than the token_probs
        returned by the other multi-mask filling functions. Use the
        all_step_seq2probs function to get something equivalent.
    'seq2probs'
        a dict mapping tuples of predicted tokens (each with length m, where m
        is the number of masked tokens in the input) to a list of m
        probabilities.
    'seq2prob'
        a dict like seq2probs, but values are the product of each token's
        predicted probabilities. We also sort this dict so the most likely
        seqs come first.
    'trie'
        maps to a Trie where each node is a predicted token (and which has an
        associated 'prob' attribute).
    """
    if not space_char.strip() and 'Ġ' in pipe.tokenizer.vocab:
        warnings.warn('Are you sure you passed the correct space_char for '
                      'fill_multi_mask_exhaustive? Looks like it should '
                      f'probably be "Ġ", not {space_char!r}.')
    text = [text]
    mask_token = pipe.tokenizer.mask_token
    xb = pipe.tokenizer(text, return_tensors='pt')
    seq_len = xb.input_ids.shape[-1]
    mask_idx = pipe.get_masked_index(xb.input_ids)[:, -1].tolist()
    all_res = []
    # Seqs is a list of lists where each nested list represents 1 predicted
    # seq and contains n_mask tuples of (tok, prob). We use it to construct
    # our trie.
    seqs = []
    n_mask = len(mask_idx)
    for i, prev_i in tqdm(zip(mask_idx, [None] + mask_idx),
                          total=n_mask, desc='Unmask'):
        with torch.no_grad():
            res = pipe.model(xb.input_ids)
        assert res.logits.shape[1] == seq_len, \
            f'Sequence length should remain {seq_len} but for mask_idx {i}, '\
            f'found seq len {res.logits.shape[1]}.'
        probs = F.softmax(res.logits[:, i, :], dim=-1)
        top_probas, top_idx = torch.topk(probs, top_k, dim=-1)
        pred_tokens = [pipe.tokenizer.convert_ids_to_tokens(row)
                       for row in top_idx]
        res = [dict(zip([pred.replace(space_char, ' ', 1) for pred in preds],
                        probas.tolist()))
               for preds, probas in zip(pred_tokens, top_probas)]
        all_res.append(res)
        if not seqs:
            # We only call this on first iteration, so we know
            # pred_tokens only has 1 row.
            seqs = [[(k, v)] for k, v in res[0].items()]
        else:
            seqs = [[*existing_seq, (k, v)]
                    for existing_seq, row in zip(seqs, res)
                    for k, v in row.items()]

        # Update token ids in batch. Originally tried to update text itself
        # and re-tokenized on each iteration but that turned out to be a bit
        # buggy: different tokenizers handle spaces differently, also have to
        # account for subwords with leading "##" (in both cases, input
        # text and output text are not formatted in the same way).
        new_input_ids = []
        for row, idx_row in zip(xb.input_ids, top_idx):
            row = row.repeat(top_k, 1)
            row[:, i] = idx_row
            new_input_ids.append(row)
        xb['input_ids'] = torch.cat(new_input_ids)
        xb['attention_mask'] = xb.attention_mask.repeat(top_k, 1)

    # Trie will help us work with sequences and probabilities later.
    T = Trie()
    for row in seqs:
        toks, probs = map(list, zip(*row))
        T.append(toks)
        for i in range(1, n_mask + 1):
            # Setting attributes this way is not ideal but in this case we
            # know
            # a. every seq we're looking for has already been added to trie
            # b. nodes are not shared between sequences with different common
            # prefixes. E.g. the leaf node in ['a', 'b', 'c'] is different
            # than the leaf node in ['x', 'y', 'c'] so settings node-level
            # attributes should be ok.
            node = T._find(toks[:i])
            node.prob = probs[i - 1]

    seq2probs = {}
    seq2prob = {}
    for seq in T:
        all_probs, total_prob = _seq2probs_from_trie(seq, T)
        key = tuple(seq)
        seq2probs[key] = all_probs
        seq2prob[key] = total_prob
    seq2prob = dict(sorted(seq2prob.items(),
                           key=lambda x: x[1], reverse=True))
    return {
        'token_probs': all_res,
        'seq2probs': seq2probs,
        'seq2prob': seq2prob,
        'trie': T
    }


def _seq2probs_from_trie(seq, T):
    """Used by fill_multi_mask_exhaustive to get the sequence of probabilities
    (and overall probability) for each sequence of predicted tokens. This is
    only used after the trie has been fully constructed.

    Parameters
    ----------
    seq: list[str]
        List of predicted tokens, where index corresponds to mask index (e.g.
        the first string should fill the first masked span).
    T: htools.Trie[list[str]]

    Returns
    -------
    tuple[list[float], float]: First item is a list of predicted probabilities
    for each token in seq at the corresponding index. Second item is the
    product of those probabilities, essentially the overall probability of this
    sequence.
    """
    # We know all sequences are the same length in this use case.
    expected_length = len(next(iter(T)))
    length = len(seq)
    if length != expected_length:
        raise ValueError(f'Invalid sequence length {length}. '
                         f'Expected {expected_length}.')
    # Use list rather than dict because we can predict the same word
    # for multiple timesteps, which results in duplicate keys in dict
    # (meaning 1 prob would be discarded).
    nodes = [getattr(T._find(seq[:i]), 'prob', 0)
             for i in range(1, length + 1)]
    return nodes, np.prod(nodes)


def next_step_seq2probs(partial_seq, res):
    """For a given (possible partial) sequence of predicted token strings,
    what are the predicted probabilities for the next step? This is mostly
    necessary because our exhaustive mask filling function output is a little
    wonky - with the auto and left to right versions, we just return the
    dict at the next index in our res['token_probs'] list.

    Parameters
    ----------
    partial_seq: list[str]
        Contains 0 or more tokens.
    res: dict
        Result of one of our multi mask filling functions.

    Returns
    -------
    dict[str, float]: Maps tokens that are likely to occur next to their
    predicted probabilities.
    """
    if 'trie' in res:
        if not partial_seq:
            node = res['trie'].head
        else:
            node = res['trie']._find(partial_seq)
        return {k: v.prob for k, v in node.edges.items()}
    return res['token_probs'][len(partial_seq)]


def all_step_seq2probs(seq, res):
    """Get a list of tok2prob dicts, 1 for each mask index when we filled a
    string with multiple masked tokens. For our auto and left to right
    functions, this literlaly just returns our token_probs list. Our exhaustive
    function output is a bit wonky so we have to do some surgery to get an
    equivalent data structure there.

    Parameters
    ----------
    seq: list[str]
        Contains m tokens, where m is the number of masked tokens in the
        corresponding input string.
    res: dict
        Result of one of our multi mask filling functions.

    Returns
    -------
    list[dict[str, float]]: Maps tokens that are likely to occur next to their
    predicted probabilities. Contains one dict for each masked token in the
    corresponding input.
    If seq was not one of our predicted seqs, we return an empty list.
    """
    if 'trie' in res:
        # Tricky bug if we don't check for seq presence: our first next_step
        # call uses an empty sequence to get the first dict of predictions, but
        # that also means we were returning probabilities for sequences that
        # didn't exist. Better to just be extra safe, and the trie makes it
        # pretty fast.
        if seq not in res['trie']:
            return []
        return [next_step_seq2probs(seq[:i], res) for i in range(len(seq))]
    return res['token_probs']


from functools import partial
from pathlib import Path
from transformers import TrainerCallback

from maminbot.generation_utils import generate


class TextGenerationCallback(TrainerCallback):
    """Generate 1 or more samples during training to give us a qualitative
    sense of different checkpoints.
    """

    def __init__(self, model, tokenizer, prompts, out_dir, **generate_kwargs):
        """
        Parameters
        ----------
        prompts: list[str]
        path: str or Path
        generate_kwargs: any
        """
        self.model = model
        self.tokenizer = tokenizer
        self.prompts = list(prompts)
        self.out_dir = out_dir
        self.out_path = Path(out_dir)/'generated_text_samples.txt'
        self.generate_kwargs = generate_kwargs
        self.generate_fn = partial(generate, out_dir=out_dir,
                                   **generate_kwargs)
        self.sep = '\n' + '=' * 79 + '\n'
        self.header_fmt = '{sep}GLOBAL STEP {global_step}{sep}'

    def on_evaluate(self, args, state, control, **kwargs):
        if not state.is_world_process_zero:
            return
        with open(self.out_path, 'a') as f:
            f.write(self.header_fmt.format(sep=self.sep,
                                           global_step=state.global_step))
        for prompt in self.prompts:
            self.generate_fn(self.model, self.tokenizer, prompt)


from datetime import datetime
from pathlib import Path
import requests
import sys

from htools.core import flatten
from maminbot import config as cfg


def load_hf_api_key(paths=('~/.huggingface', '~/secrets/.huggingface')):
    """
    Parameters
    ----------
    paths: Iterable[str]
        The paths to check for huggingface api keys. Order matters: we return
        the contents of the first non-empty file in the list.

    Returns
    -------
    str: Huggingface api key. Guaranteed to be truthy if we hit the return
    statement. Raises a FileNotFoundError if NONE of the paths exist/contain
    truthy strings.
    """
    for path in paths:
        try:
            with open(Path(path).expanduser(), 'r') as f:
                key = f.read().strip()
        except FileNotFoundError:
            key = ''
        if key: return key
    raise FileNotFoundError('Could not find any huggingface api key. '
                            f'Checked paths: {paths}')


def update_cli_command(**kwargs):
    """Used in s01 to add default args to our language model training script.
    This is helpful because adding defaults in dataclass subclasses is
    surprisingly difficult. If the user actually did specify the field, the
    defaults are not applied.

    Warning: this assumes a command like 'python train.py --user adrienne',
    or `python train.py --user=adrienne'. It is not built to handle boolean
    flags like 'python train.py --do_eval', though
    'python train.py --do_eval True' would work..

    Parameters
    ----------
    kwargs: any
    """
    sys.argv = flatten([arg.split('=') if '=' in arg else arg
                        for arg in sys.argv])
    for k, v in kwargs.items():
        if f'--{k}' not in sys.argv:
            sys.argv.extend([f'--{k}', str(v)])


def get_cli_value(field):
    """Extract the value entered at the command line by searching sys.argv.
    I.e. if user enters 'python train.py --lr .003' and the script calls
    get_cli_value('lr'), we would get .003 (as a str). If the desired field
    was not provided, we raise a ValueError.

    Parameters
    ----------
    field: str

    Returns
    -------
    str
    """
    for cur, next in zip(sys.argv, sys.argv[1:]):
        if cur == '--' + field:
            return next
    raise ValueError(f'Field {field!r} not found.')


def datetime_dirname(subdir='data/models', as_path=True, mkdir=False):
    """Return path to a new directory named like
    "~/maminbot/data/models/2022.12.31__19:49:37". In this case,
    subdir='data/models' (the default) and the date would be December 31, 2022
    and the time would be 7:49:37 pm.

    Parameters
    ----------
    subdir: str or Path
        Relative to maminbot project root.
    as_path: bool
        If True, we return a pathlib.Path obj. Str otherwise.

    Returns
    -------
    str or Path: depends on value of as_path param.
    """
    dt = datetime.now().strftime("%Y.%m.%d__%H.%M.%S")
    path = cfg.root/f'{subdir}/{dt}'
    if mkdir:
        path.mkdir()
    if not as_path:
        path = str(path)
    return path


def model_output_dirname():
    """Wrapper to get a model output dirname for s00.

    Returns
    -------
    str: Something like
    '~/maminbot/data/models/harrison/mlm/2022.12.31__18.33.19'.
    """
    user = get_cli_value("user")
    try:
        mlm = get_cli_value("mlm")
    except ValueError:
        mlm = False
    return datetime_dirname(
        f'data/models/{user}/{"mlm" if mlm else "clm"}',
        as_path=False
    )


def dataclass_to_dict(obj):
    """For cases where a dataclass or dict may be provided. This always returns
    a dict. If a dict is passed in, we make a copy to avoid mutating the
    input later.

    Parameters
    ----------
    obj: Dataclass or dict
    """
    if not isinstance(obj, dict):
        obj = getattr(obj, '__dict__', {})
    else:
        # Create a new object to avoid mutating the input.
        obj = dict(obj)
    return obj


def available_models(user, mode='', key=''):
    """Get list of available model names for a given user and (optionally)
    task.

    Parameters
    ----------
    user: str
        E.g. 'adrienne'.
    mode: str
        If non-empty, should be 'mlm' or 'clm' (for masked language model or
        causal language model).
    key: str
        Huggingface api key.

    Returns
    -------
    list[str]
    """
    mode = mode or 'M'

    def keep_fn(name):
        return mode in name

    headers = {'authorization': f'Bearer {key or load_hf_api_key()}'}
    params = {'organization': 'maminbot'}
    res = requests.get('https://huggingface.co/api/models?search=maminbot',
                       headers=headers, params=params)
    return [row['id'] for row in res.json() if
            user in row['id'] and keep_fn(row['id'])]



import os
import setuptools


def requirements(path='requirements.txt'):
    with open(path, 'r') as f:
        deps = [line.strip() for line in f]
    return deps


def version():
    path = os.path.join('maminbot', '__init__.py')
    with open(path, 'r') as f:
        for row in f:
            if not row.startswith('__version__'):
                continue
            return row.split(' = ')[-1].strip('\n').strip("'")


setuptools.setup(name='maminbot',
                 version=version(),
                 author='Harrison Mamin',
                 author_email='harrisonmamin@gmail.com',
                 description='Bots that write like members of '
                             'the Mamin family.',
                 install_requires=requirements(),
                 packages=setuptools.find_packages())




import numpy as np
from htools import load

from jabberwocky.config import C


def sample_prompt_examples(task, n=3, to_str=True):
    """Helper to dynamically construct prompts which include a numbered list.
    I think varying the examples a bit should help us get more variety in the
    completions.

    Parameters
    ----------
    task: str
        Corresponds to a path like
        ~/jabberwocky/data/dynamic_prompt_data/{task}.txt.
        The file should contain 1 prompt per line, each with a leading hyphen.
        There should typically also be a corresponding prompt like
        ~/jabberwocky/data/prompts/{task}.yaml.
    n: int
        Number of examples to sample. This should be less than the number your
        prompt requests: e.g. if you want 3 examples and 3 completions, your
        prompt would specify that the list contains 6 items (3 + 3) and you
        would pass n = 3 into this function.

    Returns
    -------
    str or list[str]: Depends on value of to_str arg. If True, we return a
    single str where each line is prefixed by a number and a singe trailing
    number is added at the end. (See below.) If False, we return a list of
    strings without adding any row numbers.

    # Example response when to_str=True (newlines have been inserted here to
    comply with pep8 conventions):
    >>> print(sample_prompt_examples('creative_writing_ideas'))

    1. Imagine that a trolley is heading towards a person tied to the tracks.
    You can pull the lever to divert it to another track, saving the person,
    but then your life savings will be destroyed. What do you do?
    2. Write a story in which a character tries to reconnect with a former
    love from a different era.
    3. You have been invited to a gala at a grand estate. Describe what you
    wear and what happens during the evening.
    4.
    """
    text = load(C.data_dir/f'dynamic_prompt_data/{task}.txt')
    rows = [row.strip('-') for row in text.splitlines() if row.strip('-')]
    sampled = list(np.random.choice(rows, n, replace=False))
    if to_str:
        sampled = [f'{i}. {row}' for i, row in enumerate(sampled, 1)]
        return '\n'.join(sampled) + f'\n{n + 1}.'
    return sampled
